use crate::Message;
use log::trace;
use std::{
    any::{Any, TypeId},
    borrow::Cow,
    collections::HashMap,
    marker::PhantomData,
};
use tokio::sync::broadcast;
use super::{BrokerError, BrokerResult};

pub trait MessageType {
    fn message_type_id(&self) -> TypeId;

    fn message_type_name(&self) -> &'static str;

    fn create_broadcast_sender(&self) -> ErasedSender;
}

pub struct BasicMessageType<T> {
    _phantom: PhantomData<T>,
}

impl<T> Default for BasicMessageType<T> {
    fn default() -> Self {
        BasicMessageType {
            _phantom: PhantomData,
        }
    }
}

impl<T: Message> MessageType for BasicMessageType<T> {
    fn message_type_id(&self) -> TypeId {
        TypeId::of::<T>()
    }

    fn message_type_name(&self) -> &'static str {
        std::any::type_name::<T>()
    }

    fn create_broadcast_sender(&self) -> ErasedSender {
        trace!(
            "Creating sender for {} ({:?})",
            std::any::type_name::<T>(),
            MessageType::type_id(self)
        );
        // TODO: configurable queue size (per message?)
        let (sender, _) = broadcast::channel::<T>(8);
        let sender: ErasedSender = Box::new(sender);
        sender
    }
}

pub type ErasedSender = Box<dyn Any + Send + Sync>;

#[derive(Debug)]
pub struct TopicEntry {
    message_type_id: TypeId,
    message_type_name: String,
    sender: ErasedSender,
}

#[derive(Debug)]
pub struct Registry {
    topics: HashMap<Cow<'static, str>, TopicEntry>,
}

impl Default for Registry {
    fn default() -> Self {
        Registry {
            topics: HashMap::new(),
        }
    }
}

impl Registry {
    fn get_erased_sender_for_type(&mut self, message_type: &dyn MessageType) -> &ErasedSender {
        let type_id = message_type.message_type_id();
        let type_name = message_type.message_type_name();
        let topic_entry = self.topics.entry(type_name.into());
        let topic_entry = topic_entry.or_insert_with(|| TopicEntry {
            message_type_id: type_id,
            message_type_name: type_name.to_string(),
            sender: message_type.create_broadcast_sender(),
        });
        &topic_entry.sender
    }

    pub fn get_sender_for_type<T: Message>(
        &mut self,
        message_type: &dyn MessageType,
    ) -> BrokerResult<&broadcast::Sender<T>> {
        let erased = self.get_erased_sender_for_type(message_type);
        (**erased)
            .downcast_ref::<broadcast::Sender<T>>()
            .ok_or(BrokerError::MismatchedType)
    }
}
