use tokio::sync::{mpsc, oneshot};

use super::{
    registry::{BasicMessageType, Registry},
    BrokerResult,
};
use crate::{Message, Publication, Subscription};

pub trait BrokerRequestInternal {
    fn run_request(self: Box<Self>, registry: &mut Registry);
}

pub trait RequestHandler {
    type Response;

    fn handle(self, registry: &mut Registry) -> Self::Response;
}

pub struct BrokerRequest<H: RequestHandler> {
    handler: H,
    response_tx: oneshot::Sender<H::Response>,
}

impl<H: RequestHandler> BrokerRequestInternal for BrokerRequest<H> {
    fn run_request(self: Box<Self>, registry: &mut Registry) {
        let response = self.handler.handle(registry);
        let _ = self.response_tx.send(response);
    }
}

impl<H: RequestHandler + Default> BrokerRequest<H> {
    fn create(response_tx: oneshot::Sender<H::Response>) -> Box<Self> {
        Box::new(Self {
            handler: H::default(),
            response_tx,
        })
    }

    pub(crate) fn new() -> (Box<Self>, oneshot::Receiver<H::Response>) {
        let (response_tx, response_rx) = oneshot::channel();
        (Self::create(response_tx), response_rx)
    }
}

pub type BrokerRequestBox = Box<dyn BrokerRequestInternal + Send + Sync>;
pub type BrokerRequestSender = mpsc::Sender<BrokerRequestBox>;

pub type SubscribeRequest<T> = BrokerRequest<Subscribe<T>>;
pub struct Subscribe<T: Message> {
    msg_type: BasicMessageType<T>,
}

impl<T: Message> Default for Subscribe<T> {
    fn default() -> Self {
        Self {
            msg_type: BasicMessageType::<T>::default(),
        }
    }
}

impl<T: Message> RequestHandler for Subscribe<T> {
    type Response = BrokerResult<Subscription<T>>;

    fn handle(self, registry: &mut Registry) -> Self::Response {
        let sender = registry.get_sender_for_type::<T>(&self.msg_type);
        sender.map(|sender| {
            let receiver = sender.subscribe();
            Subscription::new(receiver)
        })
    }
}

pub type AdvertiseRequest<T> = BrokerRequest<Advertise<T>>;
pub struct Advertise<T: Message> {
    msg_type: BasicMessageType<T>,
}

impl<T: Message> Default for Advertise<T> {
    fn default() -> Self {
        Self {
            msg_type: BasicMessageType::<T>::default(),
        }
    }
}

impl<T: Message> RequestHandler for Advertise<T> {
    type Response = BrokerResult<Publication<T>>;

    fn handle(self, registry: &mut Registry) -> Self::Response {
        let sender = registry.get_sender_for_type::<T>(&self.msg_type);
        sender.map(|sender| Publication::new(sender.clone()))
    }
}
