mod broker_task;
mod message;
mod pub_sub;

use broker_task::{
    AdvertiseRequest, BrokerError, BrokerRequestSender, BrokerResult, BrokerTask, SubscribeRequest,
};
pub use message::Message;
pub use pub_sub::*;

use futures::executor::block_on;
use tokio::sync::oneshot;

#[derive(Clone, Debug, PartialEq)]
#[non_exhaustive]
pub enum OrsbError {
    BrokerClosed,
    NoResponse,
    MismatchedType,
}

impl From<BrokerError> for OrsbError {
    fn from(err: BrokerError) -> Self {
        match err {
            BrokerError::MismatchedType => OrsbError::MismatchedType,
        }
    }
}

fn map_broker_response<T>(
    result: Result<BrokerResult<T>, oneshot::error::RecvError>,
) -> Result<T, OrsbError> {
    match result {
        Ok(Ok(value)) => Ok(value),
        Ok(Err(err)) => Err(err.into()),
        Err(_) => Err(OrsbError::NoResponse),
    }
}

#[derive(Debug, Clone)]
pub struct Orsb {
    request_tx: BrokerRequestSender,
}

impl Orsb {
    pub fn start_new() -> Self {
        Self::start_new2().0
    }

    pub(crate) fn start_new2() -> (Self, tokio::task::JoinHandle<()>) {
        let (request_tx, join_handle) = BrokerTask::start();

        let orsb = Orsb { request_tx };

        (orsb, join_handle)
    }

    pub async fn subscribe<T: Message>(&mut self) -> Result<Subscription<T>, OrsbError> {
        let (subscribe_request, response_rx) = SubscribeRequest::<T>::new();
        self.request_tx
            .send(subscribe_request)
            .await
            .or(Err(OrsbError::BrokerClosed))?;
        map_broker_response(response_rx.await)
    }

    pub async fn advertise<T: Message>(&mut self) -> Result<Publication<T>, OrsbError> {
        let (advertise_request, response_rx) = AdvertiseRequest::<T>::new();
        self.request_tx
            .send(advertise_request)
            .await
            .or(Err(OrsbError::BrokerClosed))?;
        map_broker_response(response_rx.await)
    }

    pub fn subscribe_blocking<T: Message>(&mut self) -> Result<Subscription<T>, OrsbError> {
        block_on(self.subscribe::<T>())
    }

    pub fn advertise_blocking<T: Message>(&mut self) -> Result<Publication<T>, OrsbError> {
        block_on(self.advertise::<T>())
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[derive(Clone, Debug, PartialEq)]
    struct TestMsg(u8);

    fn init_logger() {
        let _ = env_logger::try_init();
    }

    #[tokio::test]
    async fn test_sync() {
        init_logger();

        let mut orsb = Orsb::start_new();

        tokio::task::spawn_blocking(move || {
            let mut sub = orsb.subscribe_blocking::<TestMsg>().unwrap();
            let mut publ = orsb.advertise_blocking::<TestMsg>().unwrap();

            publ.publish(TestMsg(10)).unwrap();
            publ.publish(TestMsg(20)).unwrap();
            publ.publish(TestMsg(30)).unwrap();

            assert_eq!(sub.recv_blocking(), Ok(TestMsg(10)));
            assert_eq!(sub.recv_blocking(), Ok(TestMsg(20)));
            assert_eq!(sub.recv_blocking(), Ok(TestMsg(30)));
            assert_eq!(sub.try_recv(), Err(SubscriptionError::Empty));
        })
        .await
        .unwrap();
    }

    #[tokio::test]
    async fn test_async() {
        init_logger();

        let mut orsb = Orsb::start_new();

        let mut sub = orsb.subscribe::<TestMsg>().await.unwrap();
        let mut publ = orsb.advertise::<TestMsg>().await.unwrap();

        publ.publish(TestMsg(10)).unwrap();
        publ.publish(TestMsg(20)).unwrap();
        publ.publish(TestMsg(30)).unwrap();

        assert_eq!(sub.recv().await, Ok(TestMsg(10)));
        assert_eq!(sub.recv().await, Ok(TestMsg(20)));
        assert_eq!(sub.recv().await, Ok(TestMsg(30)));
        assert_eq!(sub.try_recv(), Err(SubscriptionError::Empty));
    }
}
